import sys
from enum import Enum


class RunningMode(str, Enum):
    DEV = "dev"
    PRODUCTION = "production"


def get_running_mode() -> RunningMode:
    # if uvicorn and --reload are in sys.argv, we are running in uvicorn dev mode
    _joined_argv = " ".join(sys.argv)
    if "uvicorn" in _joined_argv and "--reload" in _joined_argv:
        return RunningMode.DEV
    else:
        return RunningMode.PRODUCTION


def reactive(trigger: str | None = None):
    def decorator(func):
        func.trigger = trigger
        return func

    return decorator



from __future__ import annotations

from copy import deepcopy
from functools import partial

from pydantic import BaseModel
from pydantic.fields import ComputedFieldInfo, Field, FieldInfo


class AttrsMixin(BaseModel):
    """
    Handles the attributes of the element.
    Attributes should be defined as fields with the `Attribute` annotation.
    """

    @staticmethod
    def __define_name(k, v):
        return v.json_schema_extra.get("attribute_name") if v.json_schema_extra.get("attribute_name") else k

    def get_element_attributes(self) -> dict[str, str]:
        _without_page_ref = {
            k: v
            for k, v in self.model_fields.items()
            if v.json_schema_extra and not v.json_schema_extra.get("page_reference")
        }
        model_fields = deepcopy(_without_page_ref)
        computed_fields = {k: v for k, v in deepcopy(self.model_computed_fields).items() if k != "attrs"}
        all_fields: dict[str, FieldInfo | ComputedFieldInfo] = {**model_fields, **computed_fields}
        _attrs = {}
        for field_name, field_info in all_fields.items():
            if field_info.json_schema_extra and field_info.json_schema_extra.get("attribute"):  # type: ignore
                _attrs[self.__define_name(field_name, field_info)] = self.__getattribute__(field_name)
        return _attrs


class FactoryMixin:
    @classmethod
    def factory(cls, *args, **kwargs) -> FieldInfo:
        return Field(default_factory=partial(cls, *args, **kwargs))



from __future__ import annotations

from typing import Any, Callable, TypeVar

from lxml.etree import _Element as LxmlElement
from pydantic import PrivateAttr, computed_field

from schorle.elements.attribute import Attribute
from schorle.elements.base.base import BaseElement
from schorle.reactives.base import ReactiveBase
from schorle.reactives.classes import Classes


class Element(BaseElement):
    _base_classes: Classes = PrivateAttr(default_factory=Classes)
    classes: Classes = Classes()
    role: str | None = Attribute(default=None)
    swap: str = Attribute(default="morph", alias="hx-swap-oob")

    def __init__(self, **data):
        super().__init__(**data)
        if self.element_id is None:
            self.element_id = (
                f"schorle-{self.tag.value.lower()}-{id(self)}" if self.element_id is None else self.element_id
            )

    @computed_field(json_schema_extra={"attribute": True, "attribute_name": "hx-trigger"})  # type: ignore
    @property
    def hx_trigger(self) -> str | None:
        return ",".join(list(self.reactive_methods.keys())) if len(self.reactive_methods) > 0 else None

    @computed_field(json_schema_extra={"attribute": True, "attribute_name": "ws-send"})  # type: ignore
    @property
    def ws_send(self) -> str | None:
        return "" if len(self.reactive_methods) > 0 else None

    @computed_field()  # type: ignore
    @property
    def reactive_methods(self) -> dict[str, Callable]:
        return dict(self.get_triggers_and_methods())

    def _add_classes(self, element: LxmlElement):
        container = []
        for source in [self.classes, self._base_classes]:
            _rendered = source.render()
            if _rendered:
                container.append(_rendered)
        if container:
            element.set("class", " ".join(container))

    def get_reactive_attributes(self):
        fields = []
        for field_name in self.model_fields.keys():
            attr = getattr(self, field_name)
            if attr is not None and isinstance(attr, ReactiveBase):
                fields.append(attr)
        return fields

    def get_triggers_and_methods(self):
        for attr in dir(self):
            if (
                attr not in ["__fields__", "__fields_set__", "__signature__"]
                and attr not in self.model_computed_fields.keys()
            ):
                if callable(getattr(self, attr)):
                    method = getattr(self, attr)
                    if hasattr(method, "trigger") and method.trigger:
                        yield method.trigger, method

    def get_methods_with_attribute(self, attribute: str):
        for attr in dir(self):
            if (
                attr not in ["__fields__", "__fields_set__", "__signature__"]
                and attr not in self.model_computed_fields.keys()
            ):
                if callable(getattr(self, attr)):
                    method = getattr(self, attr)
                    if hasattr(method, attribute):
                        yield method

    async def before_render(self):
        pass

    def inject_page_reference(self, page: Any):
        for element in self.traverse():
            if isinstance(element, Element):
                for attr_name, field in element.model_fields.items():
                    if field.json_schema_extra and field.json_schema_extra.get("page_reference"):
                        setattr(element, attr_name, page)


T = TypeVar("T", bound=Element)  # todo- add strict typing for Element


class Collection(ReactiveBase[list[T]]):
    def __init__(self, value: list[T] | None = None):
        super().__init__(value=value)

    def __repr__(self):
        return f"<Collection {self._value}>"

    def __str__(self):
        return f"<Collection {self._value}>"


class Reactive(ReactiveBase[T]):
    ...



from pydantic import BaseModel

from schorle.effector import inject_effectors
from schorle.elements.base.mixins import FactoryMixin


class ReactiveModel(BaseModel, FactoryMixin, extra="allow"):
    def __init__(self, **data):
        super().__init__(**data)
        inject_effectors(self)


class ReactiveState:
    def __init__(self):
        inject_effectors(self)



import asyncio

from starlette.websockets import WebSocket

from schorle.elements.base.element import Element
from schorle.elements.page import Page
from schorle.reactives.base import ReactiveBase


class PageEmitter:
    # root emitter that controls the whole page.

    def __init__(self, page: Page):
        self.page = page
        self.emitters = self.create_emitters()

    def create_emitters(self):
        emitters = []
        for element in self.page.traverse():
            if isinstance(element, Element):
                emitters.append(ElementEmitter(element))
        return emitters

    async def emit(self, ws: WebSocket):
        emitter_tasks = []
        for emitter in self.emitters:
            emitter_tasks.append(emitter.emit_to(ws, self.page))
        await asyncio.gather(*emitter_tasks)


async def attr_to_queue(attr: ReactiveBase, target: asyncio.Queue):
    # add a single reactive attribute to the queue.
    async for value in attr:
        await target.put(value)


class ElementEmitter:
    # emitter that controls a single element.

    def __init__(self, element: Element):
        self.element = element
        self.queue = asyncio.Queue()
        self.attributes_to_queue_task = asyncio.create_task(self.attributes_to_queue())

    async def attributes_to_queue(self):
        reactive_attributes = self.element.get_reactive_attributes()
        tasks = []
        for attr in reactive_attributes:
            tasks.append(attr_to_queue(attr, self.queue))
        await asyncio.gather(*tasks)

    async def emit_to(self, ws: WebSocket, page: Page):
        while True:
            await asyncio.sleep(0.0001)  # prevent blocking
            if not self.queue.empty():
                await process_queue(self.queue, page)
                # re-render the element and send the update to the client
                await ws.send_text(self.element.render())


async def process_queue(queue: asyncio.Queue, page: Page):
    # process a queue of elements that were added to the page.
    while not queue.empty():
        element = await queue.get()
        await process_queue_output(element, page)


async def process_queue_output(element, page):
    # process a single element that was added to the page.
    if isinstance(element, list):
        for _element in element:
            await process_queue_output(_element, page)
    elif isinstance(element, Element):
        element.inject_page_reference(page)
        await element.before_render()



from uuid import uuid4

from pydantic import Field

from schorle.elements.attribute import Attribute
from schorle.elements.base.base import BaseElement
from schorle.elements.base.element import Element
from schorle.elements.page import Page
from schorle.elements.tags import HTMLTag
from schorle.theme import Theme


class Meta(BaseElement):
    tag: HTMLTag = HTMLTag.META
    charset: str = Attribute(default=None)
    name: str = Attribute(default=None)
    content: str = Attribute(default=None)


class Title(BaseElement):
    tag: HTMLTag = HTMLTag.TITLE
    text: str = "Schorle"


class Link(BaseElement):
    tag: HTMLTag = HTMLTag.LINK
    href: str = Attribute(default=None)
    rel: str = Attribute(default=None)
    type_: str | None = Attribute(default=None, alias="type")


class Script(BaseElement):
    tag: HTMLTag = HTMLTag.SCRIPT
    src: str = Attribute(default=None)
    crossorigin: str = Attribute(default=None)
    text: str = ""  # we always want to have a closing tag
    defer: str | None = Attribute(default=None)


class CSRFMeta(Meta):
    tag: HTMLTag = HTMLTag.META
    name: str = Attribute(default="schorle-csrf-token")
    content: str = Attribute(default_factory=lambda: str(uuid4()))


class Head(BaseElement):
    tag: HTMLTag = HTMLTag.HEAD
    charset_meta: Meta = Meta(charset="utf-8")
    viewport_meta: Meta = Meta(name="viewport", content="width=device-width, initial-scale=1.0")
    csrf_meta: CSRFMeta = CSRFMeta()
    dev_meta: Meta | None = None
    title: Title = Title()
    favicon: Link = Link(href="/favicon.svg", rel="icon", type_="image/svg+xml")
    # css-related
    daisy_ui: Link = Link(
        href="https://cdn.jsdelivr.net/npm/daisyui@4.4.22/dist/full.min.css", rel="stylesheet", type_="text/css"
    )
    tailwind: Script = Script(src="https://cdn.tailwindcss.com")
    # htmx
    htmx: Script = Script(src="https://unpkg.com/htmx.org@1.9.10", crossorigin="anonymous")
    htmx_ws: Script = Script(src="https://unpkg.com/htmx.org@1.9.10/dist/ext/ws.js")
    idiomorph: Script = Script(src="https://unpkg.com/idiomorph@0.3.0")
    # todo: use idiomorph-htmx when bug is fixed
    # idiomorph_htmx: Script.provide(src="https://unpkg.com/idiomorph/dist/idiomorph-ext.min.js")
    # client-side bundle
    bundle: Script = Script(src="/_schorle/assets/bundle.js", crossorigin="anonymous")


class Footer(BaseElement):
    tag: HTMLTag = HTMLTag.FOOTER
    text: str = ""


class DevFooter(BaseElement):
    tag: HTMLTag = HTMLTag.FOOTER
    text: str = ""


class Body(BaseElement):
    tag: HTMLTag = HTMLTag.BODY
    footer: DevFooter = DevFooter(element_id="schorle-footer")


class Html(BaseElement):
    tag: HTMLTag = HTMLTag.HTML
    theme: Theme = Attribute(..., alias="data-theme")
    head: Head = Field(default_factory=Head)
    body: Body = Field(default_factory=Body)


class Div(Element):
    tag: HTMLTag = HTMLTag.DIV


class EventHandler(BaseElement):
    tag: HTMLTag = HTMLTag.DIV
    element_id: str = "schorle-event-handler"
    hx_ws: str = Attribute(default="ws,event-header", alias="hx-ext")
    ws_connect: str = Attribute(default="/_schorle/events", alias="ws-connect")
    content: Page


class MorphWrapper(BaseElement):
    tag: HTMLTag = HTMLTag.DIV
    element_id: str = "schorle-morph-wrapper"
    morph: str = Attribute(default="morph", alias="hx-ext")
    handler: EventHandler


class BodyWithPage(Body):
    wrapper: MorphWrapper


class Paragraph(Element):
    tag: HTMLTag = HTMLTag.P


class Span(Element):
    tag: HTMLTag = HTMLTag.SPAN



from __future__ import annotations

from schorle.app import Schorle
from schorle.effector import effector
from schorle.elements.button import Button
from schorle.elements.html import Div
from schorle.elements.page import Page, PageReference
from schorle.reactives.classes import Classes
from schorle.reactives.state import ReactiveModel
from schorle.reactives.text import Text

app = Schorle()


class Counter(ReactiveModel):
    value: int = 0

    @effector
    async def increment(self):
        self.value += 1

    @effector
    async def decrement(self):
        self.value -= 1


class DecrementButton(Button):
    text: Text = Text("Decrement")
    page: PageWithButton = PageReference()
    classes: Classes = Classes("btn-error")

    async def before_render(self):
        await self.page.counter.decrement.subscribe(self._switch_off, trigger=True)
        await self.page.counter.increment.subscribe(self._switch_off, trigger=True)

    async def _switch_off(self, counter: Counter):
        if counter.value <= 0:
            await self.classes.append("btn-disabled")
        else:
            await self.classes.remove("btn-disabled")


class Buttons(Div):
    classes: Classes = Classes("space-x-4 flex flex-row justify-center items-center")
    increment: Button = Button.factory(text=Text("Increment"), classes=Classes("btn-success"))
    decrement: DecrementButton = DecrementButton.factory()
    page: PageWithButton = PageReference()

    async def before_render(self):
        self.increment.add_callback("click", self.page.counter.increment)
        self.decrement.add_callback("click", self.page.counter.decrement)


class CounterView(Div):
    page: PageWithButton = PageReference()

    async def update(self, counter: Counter):
        await self.text.update(f"Clicked {counter.value} times")

    async def before_render(self):
        await self.page.counter.increment.subscribe(self.update, trigger=True)
        await self.page.counter.decrement.subscribe(self.update, trigger=True)


class PageWithButton(Page):
    counter: Counter = Counter.factory()
    classes: Classes = Classes("space-y-4 flex flex-col justify-center items-center h-screen w-screen")
    buttons: Buttons = Buttons.factory()
    counter_view: CounterView = CounterView.factory()


@app.get("/")
def get_page():
    return PageWithButton()



from __future__ import annotations

from pydantic import Field

from schorle.app import Schorle
from schorle.effector import effector
from schorle.elements.base.element import Collection, Reactive
from schorle.elements.button import Button
from schorle.elements.html import Div
from schorle.elements.inputs import Input
from schorle.elements.page import Page, PageReference
from schorle.reactives.classes import Classes
from schorle.reactives.state import ReactiveModel
from schorle.reactives.text import Text

app = Schorle()


class TodoList(ReactiveModel):
    items: list[str] = Field(default_factory=lambda: ["Buy milk", "Buy eggs", "Buy bread"])

    @effector
    async def add_item(self, item: str):
        self.items.append(item)

    @effector
    async def remove_item(self, item: str):
        self.items.remove(item)


class InputSection(Div):
    classes: Classes = Classes("flex flex-row justify-center items-center w-6/12 space-x-4")
    input_text: Input = Input(placeholder="Enter text here")
    add_button: Button = Button(text="Add", classes=Classes("btn-primary"))
    page: TodoPage = PageReference()

    async def before_render(self):
        self.add_button.add_callback("click", self._on_click)

    async def _on_click(self):
        new_input = self.input_text.value.get()
        if new_input:
            await self.page.todo_list.add_item(new_input)
            await self.input_text.clear()


class TodoItem(Div):
    classes: Classes = Classes("w-full flex flex-row justify-between items-center")
    remove_button: Button = Button.factory(text=Text("Remove"), classes=Classes("btn-error"))
    page: TodoPage = PageReference()

    async def before_render(self):
        self.remove_button.add_callback("click", self._on_click)

    async def _on_click(self):
        await self.page.todo_list.remove_item(self.text.get())


class TodoView(Div):
    classes: Classes = Classes("flex w-96 flex-col space-y-4 p-4")
    headline: Reactive[Div] = Reactive.factory()
    todo_items: Collection[TodoItem] = Collection.factory()
    page: TodoPage = PageReference()

    async def update_items(self, todo_list: TodoList):
        new_items = [TodoItem(text=Text(item)) for item in todo_list.items]
        await self.todo_items.update(new_items)

    async def update_headline(self, todo_list: TodoList):
        _text = (
            f"Todo List with {len(todo_list.items)} item{'s' if len(todo_list.items) > 1 else ''}"
            if todo_list.items
            else "Todo List is empty"
        )
        await self.headline.update(Div(text=Text(_text), classes=Classes("text-2xl")))

    async def before_render(self):
        for updater in [self.update_items, self.update_headline]:
            await self.page.todo_list.add_item.subscribe(updater, trigger=True)
            await self.page.todo_list.remove_item.subscribe(updater, trigger=True)


class TodoPage(Page):
    todo_list: TodoList = TodoList.factory()
    classes: Classes = Classes("flex flex-col justify-center items-center h-screen w-screen")
    input_section: InputSection = InputSection.factory()
    todo_view: TodoView = TodoView.factory()


@app.get("/")
def get_page():
    return TodoPage()



from typing import List

from pydantic import Field

from schorle.elements.attribute import Attribute
from schorle.elements.html import Div
from schorle.elements.inputs import Input
from schorle.reactives.base import ReactiveBase
from schorle.reactives.classes import Classes


class Tab(Div):
    render_behaviour: str = "flatten"
    input_field: Input = Field(default_factory=lambda: Input(input_type="radio", classes=Classes("tab"), role="tab"))
    content: Div = Field(default_factory=lambda: Div(classes=Classes("tab-content"), role="tabpanel"))


class TabsList(Div):
    role: str = Attribute("tablist")
    children: ReactiveBase[List[Tab]] = Field(default_factory=ReactiveBase)



from __future__ import annotations

from schorle.app import Schorle
from schorle.effector import effector
from schorle.elements.button import Button
from schorle.elements.page import Page
from schorle.reactives.classes import Classes
from schorle.reactives.state import ReactiveModel
from schorle.reactives.text import Text
from schorle.utils import reactive

app = Schorle()


class Counter(ReactiveModel):
    value: int = 0

    @effector
    async def increment(self):
        self.value += 1


class ButtonWithCounter(Button):
    text: Text = Text("Click me!")
    counter: Counter = Counter.factory()

    @reactive("click")
    async def handle(self):
        await self.counter.increment()

    async def _on_increment(self, counter: Counter):
        await self.text.update(f"Clicked {counter.value} times")
        await self.classes.toggle("btn-success")

    async def before_render(self):
        await self.counter.increment.subscribe(self._on_increment)


class PageWithButton(Page):
    classes: Classes = Classes("flex flex-col justify-center items-center h-screen w-screen")
    first_button: ButtonWithCounter = ButtonWithCounter.factory()
    second_button: ButtonWithCounter = ButtonWithCounter.factory()


@app.get("/")
def get_page():
    return PageWithButton()



from pydantic import Field
from pydantic.fields import computed_field

from schorle.elements.attribute import Attribute
from schorle.elements.base.element import Element
from schorle.elements.tags import HTMLTag
from schorle.reactives.base import ReactiveBase
from schorle.reactives.classes import Classes
from schorle.utils import reactive


class Input(Element):
    tag: HTMLTag = HTMLTag.INPUT
    value: ReactiveBase[str] = Field(default=ReactiveBase(""), description="Value of the input field")
    _base_classes: Classes = Classes("input", "form-control")
    hx_include: str = Attribute(default="this", alias="hx-include", private=True)
    placeholder: str | None = Attribute(default=None)
    name: str = Attribute(default="default")
    input_type: str = Attribute(default="text", alias="type")

    def __init__(self, **data):
        super().__init__(**data)
        self.name = self.element_id

    async def clear(self):
        await self.value.update("")

    @reactive("change")
    async def on_change(self, new_value: str):
        await self.value.update(new_value, skip_render=True)


class Slider(Input):
    _base_classes: Classes = Classes("range", "form-control")
    input_type: str = Attribute(default="range", alias="type")
    minimum: int = 0
    maximum: int = 100

    @computed_field(json_schema_extra={"attribute": True, "attribute_name": "min"})  # type: ignore
    @property
    def min_attr(self) -> str | None:
        return str(self.minimum)

    @computed_field(json_schema_extra={"attribute": True, "attribute_name": "max"})  # type: ignore
    @property
    def max_attr(self) -> str | None:
        return str(self.maximum)



from __future__ import annotations

from asyncio import Queue
from typing import AsyncIterator, Generic, TypeVar

from loguru import logger
from pydantic import BaseModel, PrivateAttr

from schorle.elements.base.mixins import FactoryMixin

T = TypeVar("T")


class ReactiveBase(BaseModel, Generic[T], FactoryMixin):
    _render_queue: Queue = PrivateAttr(default_factory=Queue)
    _value: T | None = PrivateAttr()

    async def update(self, value: T | None, *, skip_render: bool = False):
        self._value = value
        logger.debug(f"Updating {self} with {value}")
        if not skip_render:
            await self._render_queue.put(self)
        else:
            logger.debug(f"Skipping render for {self}")

    async def __aiter__(self) -> AsyncIterator[T | None]:
        while True:
            await self._render_queue.get()
            yield self._value

    def get(self) -> T | None:
        return self._value

    def __init__(self, value: T | None = None):
        super().__init__()
        self._value = value

    def __repr__(self):
        return f"<{self.__class__.__name__} {self._value}>"

    def __str__(self):
        return f"<{self.__class__.__name__} {self._value}>"
